(*
 * Copyright (c) 2009, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import sepalg.

Set Implicit Arguments.

Record age_sepalg_facts (A:Type) (sa:sepalg A) (age:A -> A -> Prop) (unage: A -> A) :=
mk_ASA_facts
{ asf_wf : well_founded (fun x y => age y x)
; asf_unage_identity: forall x z, age x z -> identity z -> identity x
; asf_unage : forall x, age (unage x) x
; asf_commute1 : forall x y z x', join x y z -> age x x' ->
    exists y':A, exists z':A, join x' y' z' /\ age y y' /\ age z z'
; asf_commute2 : forall x y z z', join x y z -> age z z' ->
    exists x':A, exists y':A, join x' y' z' /\ age x x' /\ age y y'
; asf_commute3 : forall x x' y' z', join x' y' z' -> age x x' ->
    exists y:A, exists z:A, join x y z /\ age y y' /\ age z z'
; asf_commute4 : forall z x' y' z', join x' y' z' -> age z z' ->
    exists x:A, exists y:A, join x y z /\ age x x' /\ age y y'
}.

Class ASA `(as_alg : sepalg A) :=
mkASA
{ age1 : A -> option A
; unage : A -> A
; age := fun x y => age1 x = Some y
; as_facts : age_sepalg_facts as_alg age unage
}.

Definition age1_join  `{ASA A}: forall x y z x' : A,
       join x y z ->
       age x x' ->
       exists y' : A,
         exists z' : A, join x' y' z' /\ age y y' /\ age z z'
       := asf_commute1 as_facts.

Definition age1_join2  `{ASA A}: forall x y z z' : A,
       join x y z ->
       age z z' ->
       exists x' : A, exists y' : A, join x' y' z' /\ age x x' /\ age y y'
      := asf_commute2 as_facts.

Definition unage1_join  `{ASA A}: forall x x' y' z' : A,
       join x' y' z' ->
       age x x' ->
       exists y : A,
         exists z : A, join x y z /\ age y y' /\ age z z'
       := asf_commute3 as_facts.

Definition unage1_join2  `{ASA A}: forall z x' y' z' : A,
       join x' y' z' ->
       age z z' ->
       exists x : A, exists y : A, join x y z /\ age x x' /\ age y y'
      := asf_commute4 as_facts.


Lemma age1_None_joins `{ASA A}: forall phi1 phi2, joins phi1 phi2 -> age1 phi1 = None -> age1 phi2 = None.
Proof.
intros A HA XA.
intros.
destruct H.
case_eq (age1 phi2); intros; auto.
destruct (age1_join _ _ (join_com _ _ _ H) H1) as [phi1' [x' [? [? ?]]]].
unfold age in *; rewrite H0 in H3; inv H3.
Qed.

Lemma age1_joins_eq `{ASA A}: forall phi1 phi2,
        joins phi1 phi2 -> 
        forall phi1', age1 phi1 = Some phi1' -> 
        forall phi2', age1 phi2 = Some phi2' ->
        joins phi1' phi2'.
Proof.
intros A HA XA.
intros.
destruct H.
destruct (age1_join _ _ H H0) as [phi7 [x' [? [? ?]]]].
unfold age in *; rewrite H1 in H3; inv H3.
exists x'; auto.
Qed.

Definition age_induction `{ASA A} := well_founded_induction (asf_wf as_facts).

Require Import sepalg_generators.

Section BIJECTION.
Variable A B : Type.
Variable sa: sepalg A.
Variable bijAB: bijection A B.
Variable asa : ASA sa.

Definition age1B  (x: B) : option B :=
     match age1 (bij_g _ _ bijAB x) with
     | Some y => Some (bij_f _ _ bijAB y)
     | None => None
     end.

Definition ageB (x y: B) :=age1B x = Some y.

Definition unageB (x:B) : B :=
  bij_f _ _ bijAB (unage (bij_g _ _ bijAB x)).

Lemma asa_bij_wellfounded:
    well_founded (fun x y => ageB y x).
Proof.
intros.
generalize (asf_wf as_facts); intro.
unfold ageB, age1B.
destruct bijAB as [f g fg gf].
simpl in *.
intro b.
spec H (g b).
remember (g b) as a.
revert b Heqa; induction H; intros.
constructor;  intros.
revert H1; case_eq (age1 (g b)); intros; inv H2.
spec H0 a H1 (f a).
spec H0; auto.
Qed.

Definition saB := (sa_bijection _ _ bijAB sa).
Existing Instance saB.

Lemma asa_bij_unage_identity:
   forall x z : B, ageB x z -> identity z -> identity x.
Proof.
unfold ageB, age1B.
unfold saB.
destruct bijAB as [f g fg gf].
simpl.
intros x z.
case_eq (age1 (g x)); intros; try discriminate.
inv H0.
rewrite identity_unit_equiv.
simpl.
cut (identity (g x)).
rewrite identity_unit_equiv; auto.
apply (asf_unage_identity as_facts) with a; auto.
rewrite identity_unit_equiv in H1.
red in H1.
simpl in H1.
rewrite gf in H1.
rewrite identity_unit_equiv; auto.
Qed.

Lemma asa_bij_unage : forall x, ageB (unageB x) x.
Proof.
  intros.
  unfold ageB, age1B, unageB.
  rewrite bij_gf.
  rewrite (asf_unage as_facts).
  rewrite bij_fg.
  auto.
Qed.

Definition asa_bijection:  ASA (sa_bijection _ _ bijAB sa).
intros.
apply mkASA with (fun (x: B) =>
                                 match age1 (bij_g _ _ bijAB x) with
                                 | Some y => Some (bij_f _ _ bijAB y)
                                 | None => None
                             end) unageB.
constructor.
(* well_founded *)
exact asa_bij_wellfounded.
exact asa_bij_unage_identity.
exact asa_bij_unage.

(* commute1 *)
unfold ageB, age1B, saB.
destruct bijAB as [f g fg gf].
simpl in *.
intros.
revert H0; case_eq (age1 (g x)); intros; try discriminate.
inv H1.
rename a into gx'.
destruct (asf_commute1 as_facts _ _ _ _ H H0) as  [gy' [gz' [? [? ?]]]].
exists (f gy'); exists (f gz').
repeat rewrite gf.
split; auto.
rewrite H2; rewrite H3.
auto.

(* commute2 *)
unfold ageB, age1B, saB.
destruct bijAB as [f g fg gf].
simpl in *.
intros.
revert H0; case_eq (age1 (g z)); intros; try discriminate.
inv H1.
rename a into gz'.
destruct (asf_commute2 as_facts _ _ _ _ H H0) as  [gx' [gy' [? [? ?]]]].
exists (f gx'); exists (f gy').
repeat rewrite gf.
split; auto.
rewrite H2; rewrite H3.
auto.

(* commute3 *)
unfold ageB, age1B, saB.
destruct bijAB as [f g fg gf].
simpl in *.
intros.
revert H0; case_eq (age1 (g x)); intros; try discriminate.
inv H1.
rename a into gx'.
rewrite gf in *.
destruct (asf_commute3 as_facts _ _ _ _ H H0) as  [gy' [gz' [? [? ?]]]].
exists (f gy'); exists (f gz').
repeat rewrite gf.
split; auto.
rewrite H2; rewrite H3.
repeat rewrite fg; split; auto.

(* commute4 *)
unfold ageB, age1B, saB.
destruct bijAB as [f g fg gf].
simpl in *.
intros.
revert H0; case_eq (age1 (g z)); intros; try discriminate.
inv H1.
rename a into gz'.
rewrite gf in *.
destruct (asf_commute4 as_facts _ _ _ _ H H0) as  [gx' [gy' [? [? ?]]]].
exists (f gx'); exists (f gy').
repeat rewrite gf.
split; auto.
rewrite H2; rewrite H3.
repeat rewrite fg; split; auto.
Defined.

End BIJECTION.

Section PROD.

Variable A B : Type.
Variable saA: sepalg A.
Variable saB: sepalg B.
Variable asaA: ASA saA.

Definition asa_prod: ASA (sa_prod saA saB).
intros.
apply mkASA with (fun x : A*B => match age1 (fst x) with
                                                 | Some y => Some (y, snd x)
                                                 | None => None
                                                 end)
                 (fun x : A*B => (unage (fst x), snd x)).
constructor.

(*  well_founded *)
generalize (asf_wf as_facts); intro.
unfold age in H.
intros [a b].
spec H a.
induction H; intros.
constructor.
intros.
simpl in *.
revert H1; case_eq (age1 x); intros; inv H2.
auto.

(* unage_identity *)
intros.
revert H; case_eq (age1 (fst x)) ;intros; try discriminate.
inv H1.
destruct x; simpl in *.
rewrite identity_unit_equiv in H0.
destruct H0; simpl in *.
rewrite identity_unit_equiv; split; simpl; auto.
cut (identity a0).
rewrite identity_unit_equiv; auto.
apply (asf_unage_identity as_facts) with a; auto.
rewrite identity_unit_equiv; auto.

(* unage *)
simpl; intros.
destruct x; simpl.
generalize (asf_unage as_facts a).
unfold age.
intro; rewrite H; auto.

(* commute1 *)
intros [xa xb] [ya yb] [za zb] [xa' xb'] [? ?].
simpl in *.
case_eq (age1 xa); intros; inv H2.
destruct (asf_commute1 as_facts _ _ _ _ H H1) as [ya' [za' [? [? ?]]]].
exists (ya',yb); exists (za',zb); 
rewrite H3; rewrite H4; repeat split; auto.

(* commute2 *)
intros [xa xb] [ya yb] [za zb] [za' zb'] [? ?].
simpl in *.
case_eq (age1 za); intros; inv H2.
destruct (asf_commute2 as_facts _ _ _ _ H H1) as [xa' [ya' [? [? ?]]]].
exists (xa',xb); exists (ya',yb); 
rewrite H3; rewrite H4; repeat split; auto.

(* commute3 *)
intros [xa xb] [xa' xb'] [ya' yb'] [za' zb'] [? ?].
simpl in *.
case_eq (age1 xa); intros; inv H2.
destruct (asf_commute3 as_facts _ _ _ _ H H1) as [ya [za [? [? ?]]]].
exists (ya,yb'); exists (za,zb'); simpl.
rewrite H3; rewrite H4; repeat split; auto.

(* commute4 *)
intros [za zb] [xa' xb'] [ya' yb'] [za' zb'] [? ?].
simpl in *.
case_eq (age1 za); intros; inv H2.
destruct (asf_commute4 as_facts _ _ _ _ H H1) as [xa [ya [? [? ?]]]].
exists (xa,xb'); exists (ya,yb'); simpl.
rewrite H3; rewrite H4; repeat split; auto.
Defined.

End PROD.


Inductive fashionable `{ASA A} : A -> A -> Prop :=
 | intro_fashion_none : forall x y,
       age1 x = None ->
       age1 y = None ->
       fashionable x y
 | intro_fashion_some : forall x x' y y',
       age1 x = Some x' ->
       age1 y = Some y' ->
       fashionable x' y' ->
       fashionable x y.

Lemma fashionable_refl `{ASA A} : reflexive _ fashionable.
Proof.
  repeat intro.
  induction x using (well_founded_induction (asf_wf as_facts)).
  case_eq (age1 x); intros.
  apply intro_fashion_some with a a; auto.
  apply intro_fashion_none; auto.
Qed.

Lemma fashionable_trans `{ASA A} : transitive _  fashionable.
Proof.
  intros.
  intro x.
  induction x using (well_founded_induction (asf_wf as_facts)).
  intros.
  inv H1; inv H2;
    rewrite H4 in H1; inv H1.
  apply intro_fashion_none; auto.
  apply intro_fashion_some with x' y'0; auto.
  eapply H0; eauto.
Qed.

Lemma fasionable_sym `{ASA A} : symmetric _ fashionable.
Proof.
  intros; intro x.
  induction x using (well_founded_induction (asf_wf as_facts)).
  intros.
  inv H1.
  apply intro_fashion_none; auto.
  apply intro_fashion_some with y' x'; auto.
Qed.

Lemma comparable_fashionable `{ASA A} :
  inclusion A comparable fashionable.
Proof.
  intros; intro.
  induction x using age_induction; intros.
  apply comparable_common_unit in H1.
  destruct H1 as [u [? ?]].
  case_eq (age1 x); intros.
  destruct (age1_join2 _ _ H1 H3) as [p [q [? [? ?]]]].
  assert (q = a) by congruence; subst q.
  destruct (age1_join _ _ H2 H5)
    as [s [t [? [? ?]]]].
  assert (s = t) by congruence; subst t.
  econstructor 2; eauto.
  apply H0; auto.
  apply common_unit_comparable.
  exists p; split; auto.
  case_eq (age1 y); intros.
  destruct (age1_join2 _ _ H2 H4) as [p [q [? [? ?]]]].
  destruct (age1_join _ _ H1 H6) as [s [t [? [? ?]]]].
  hnf in H9.
  rewrite H3 in H9; discriminate.
  constructor 1; auto.
Qed.
