(*
 * Copyright (c) 2009, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import knot.

Require Import Wellfounded.

Module Knot_Lemmas (K : KNOT).
  Import K.TF.
  Import K.
  
  Lemma well_founded_age : well_founded (transp knot knot_age).
  Proof.
    intro x.
    set (n := level x).
    generalize (refl_equal n); unfold n at 2.
    clearbody n.
    revert x; induction n; intros.
    constructor; intros.
    unfold transp, knot_age, knot_age1 in H0.
    unfold level in H.
    destruct (unsquash x).
    destruct n; discriminate.
    constructor; intros.
    apply IHn.
    unfold transp, knot_age, knot_age1, level in *.
    destruct (unsquash x); destruct n0; try discriminate.
    inv H0.
    rewrite unsquash_squash; simpl in *.
    inv H; auto.
  Qed.

  Lemma unsquash_inj : forall k1 k2,
    unsquash k1 = unsquash k2 ->
    k1 = k2.
  Proof.
    intros.
    rewrite <- (squash_unsquash k1).
    rewrite <- (squash_unsquash k2).
    rewrite H.
    trivial.
  Qed.
  Implicit Arguments unsquash_inj.

  Lemma squash_surj : forall k, exists n, exists Fp,
    squash (n, Fp) = k.
  Proof.
  intros.
  remember (unsquash k).
  destruct p.
  exists n.
  exists f.
  rewrite Heqp.
  rewrite squash_unsquash.
  trivial.
  Qed.

  Lemma unsquash_approx : forall k n Fp,
    unsquash k = (n, Fp) ->
    Fp = fmap (approx n) Fp.
  Proof.
    intros.
    generalize H; intro.
    rewrite <- (squash_unsquash k) in H.
    rewrite H0 in H.
    rewrite unsquash_squash in H.
    inversion H.
    rewrite H2.
    symmetry.
    trivial.
  Qed.
  Implicit Arguments unsquash_approx.
  
  Lemma approx_approx1 : forall m n,
    approx n = approx n oo approx (m+n).
  Proof.
    intros.
    extensionality p x; destruct x as [k o].
    unfold approx, compose; simpl.
    destruct (le_gt_dec n (level k)); auto.
    destruct (le_gt_dec (m+n) (level k)); auto.
    elimtype False; omega.
  Qed.

  Lemma approx_approx2 : forall m n,
    approx n = approx (m+n) oo approx n.
  Proof.
    intros.
    extensionality p x; destruct x as [k o].
    unfold approx, compose; simpl.
    destruct (le_gt_dec (m+n) (level k)); auto.
    destruct (le_gt_dec n (level k)); auto.
    elimtype False; omega.
  Qed.

  (* These are provided since sometimes it is tedious to break things out;
      they are not interesting except as engineering artifacts. *)
  Lemma unsquash_squash_unfolded : forall nf,
    unsquash (squash nf) = (fst nf, fmap (approx (fst nf)) (snd nf)).
  Proof.
    intros.
    destruct nf.
    apply unsquash_squash.
  Qed.
	
  Lemma unsquash_approx_unfolded : forall k,
    unsquash k = (fst (unsquash k), fmap (approx (fst (unsquash k))) (snd (unsquash k))).
  Proof.
    intros.
    case_eq (unsquash k); intros.
    simpl.
    apply injective_projections; simpl; trivial.
    apply (unsquash_approx H).
  Qed.

  Lemma knot_unage_correct : forall k, knot_age (knot_unage k) k.
  Proof.
    intros; unfold knot_unage, knot_age, knot_age1.
    case_eq (unsquash k); intros.
    rewrite unsquash_squash; simpl.
    f_equal.
    apply unsquash_inj.
    rewrite unsquash_squash.
    rewrite H.
    f_equal.
    change (fmap (approx n) (fmap (approx (S n)) f))
      with ((fmap (approx n) oo (fmap (approx (S n)))) f).
    rewrite fmap_comp.
    rewrite <- (approx_approx1 1).
    rewrite <- (unsquash_approx H); auto.
  Qed.

  (** [squash] and [unsquash] _almost_ form a galois connection.
     The almost is because the order given to [nat * F predicate]
     is only a preorder, and not a partial order.  One could achieve
     a partial order by quotenting [nat * F predicate] by the
     congruence generated by the order, achieving a true galois
     connection on this quotiented set.  The quotenting would
     put together into an equivalance class all [(n,f)] and [(n,f')]
     such that [f] and [f'] are the same when approximated to [n].
   *)
  Section galois_connection.
    Definition k_le := transp knot (clos_refl_trans _ knot_age).
    
    Lemma k_le_eq_or_trans : forall x y,
      k_le x y -> x = y \/ transp knot (clos_trans _ knot_age) x y.
    Proof.
      intros.
      induction H; auto.
      right; hnf.
      apply t_step; auto.
      destruct IHclos_refl_trans1; subst; auto.
      destruct IHclos_refl_trans2; subst; auto.
      right.
      unfold transp in *.
      apply t_trans with y; auto.
    Qed.

    Lemma k_le_refl : forall x, k_le x x.
    Proof.
      intros; apply rt_refl.
    Qed.

    Lemma k_le_trans : forall x y z,
      k_le x y -> k_le y z -> k_le x z.
    Proof.
      intros; eapply rt_trans; eauto.
    Qed.

    Lemma transpose_clos_trans : forall A R x y,
      clos_trans A (transp A R) x y <-> transp A (clos_trans A R) x y.
    Proof.
      unfold transp; intuition.
      
      elim H; intros.
      apply t_step; auto.
      apply t_trans with y0; auto.
      
      elim H; intros.
      apply t_step; auto.
      apply t_trans with y0; auto.
    Qed.

    Lemma k_le_antisym : forall x y,
      k_le x y -> k_le y x -> x = y.
    Proof.
      intros x y H H1.
      apply k_le_eq_or_trans in H.
      apply k_le_eq_or_trans in H1.
      intuition.
      unfold transp in *.
      elimtype False.
      assert (transp _ (clos_trans knot knot_age) x x).
      apply t_trans with y; auto.
      rewrite <- transpose_clos_trans in H1.
      clear -H1.
      induction x using
        (well_founded_induction (wf_clos_trans _ _ (well_founded_age))).
      apply H with x; auto.
    Qed.

    (* Note this is only a _preorder_ *)
    Definition nF_le (x:nat * F predicate) (y:nat * F predicate) : Prop :=
      let (n,f) := x in
      let (n',f') := y in
      n <= n' /\ fmap (approx n) f = fmap (approx n) f'.

    Lemma nF_le_refl : forall x, nF_le x x.
    Proof.
      unfold nF_le.
      intros; intuition.
    Qed.

    Lemma nF_le_trans : forall x y z,
      nF_le x y -> nF_le y z -> nF_le x z.
    Proof.
      intros; destruct x; destruct y; destruct z.
      simpl in *; intuition.
      rewrite H2.
      assert (exists m, n0 = m+n).
      exists (n0 - n).
      omega.
      destruct H0.
      subst n0.
      rewrite (approx_approx1 x n).
      rewrite <- fmap_comp.
      unfold compose at 1.
      rewrite H3.
      auto.
    Qed.

    Theorem galois_connection : forall k n f,
      k_le k (squash (n,f)) <-> nF_le (unsquash k) (n,f).
    Proof.
      intros; split; intro H.
      remember (squash (n,f)) as k'.
      revert n f Heqk'.
      induction H; intros; subst.
      unfold knot_age, knot_age1 in H.
      rewrite unsquash_squash in H.
      destruct n; try discriminate.
      inv H.
      rewrite unsquash_squash.
      hnf; split; auto.
      change (fmap (approx n) (fmap (approx (S n)) f)) with
        ((fmap (approx n) oo fmap (approx (1 + n))) f).
      rewrite fmap_comp.
      rewrite <- (approx_approx1 1).
      change (fmap (approx n) (fmap (approx n) f)) with
        ((fmap (approx n) oo fmap (approx (0 + n))) f).
      rewrite fmap_comp.
      rewrite <- (approx_approx1 0).
      auto.

      hnf.
      rewrite unsquash_squash.
      split; auto.
      change (fmap (approx n) (fmap (approx n) f)) with
        ((fmap (approx n) oo fmap (approx (0 + n))) f).
      rewrite fmap_comp.
      rewrite <- (approx_approx1 0).
      auto.

      case_eq (unsquash y); intros.
      apply nF_le_trans with (n0,f0).
      apply IHclos_refl_trans2.
      apply unsquash_inj.
      rewrite unsquash_squash.
      rewrite H1.
      f_equal.
      apply (unsquash_approx H1).
      rewrite <- H1.
      apply IHclos_refl_trans1; auto.      

      case_eq (unsquash k); intros.
      rewrite H0 in H.
      destruct H.
      assert (exists m, n = m + n0).
      exists (n - n0); omega.
      destruct H2; subst n.
      clear H.
      revert k f n0 f0 H1 H0.
      induction x; intros.
      simpl.
      replace (squash (n0,f)) with k.
      apply rt_refl.
      apply unsquash_inj.
      rewrite unsquash_squash.
      rewrite H0.
      f_equal.
      rewrite <- H1.
      apply (unsquash_approx H0).
      
      apply rt_trans with (squash (x+n0,f)).
      2: apply IHx with f0; auto.
      apply rt_step.
      unfold knot_age, knot_age1.
      rewrite unsquash_squash.
      simpl.
      f_equal.
      apply unsquash_inj.
      repeat rewrite unsquash_squash.
      f_equal.
      remember (x+n0) as z.
      change (fmap (approx z) (fmap (approx (S z)) f))
        with ((fmap (approx z) oo (fmap (approx (1+z)))) f).
      rewrite fmap_comp.
      rewrite <- approx_approx1.
      auto.
    Qed.

  End galois_connection.

End Knot_Lemmas.
